@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2000-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2000-2013 Free Software Foundation, Inc.
+ * Copyright (C) 2012,2013 Nikos Mavrogiannopoulos
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -7,7 +8,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -46,9 +47,11 @@
 #include "gnutls_datum.h"
 #include "gnutls_constate.h"
 #include "ext/max_record.h"
+#include <ext/heartbeat.h>
 #include <gnutls_state.h>
 #include <gnutls_dtls.h>
 #include <gnutls_dh.h>
+#include <random.h>
 
 struct tls_record_st {
   uint16_t header_size;
@@ -63,7 +66,7 @@ struct tls_record_st {
 };
 
 /**
- * gnutls_record_disable_padding:
+ * gnutls_record_disable_padding:  
  * @session: is a #gnutls_session_t structure.
  *
  * Used to disabled padding in TLS 1.0 and above.  Normally you do not
@@ -71,13 +74,33 @@ struct tls_record_st {
  * complain if a server pads the encrypted data.  This of course will
  * disable protection against statistical attacks on the data.
  *
- * Normally only servers that require maximum compatibility with everything
- * out there, need to call this function.
+ * This functions is defunt since 3.1.7. Random padding is disabled
+ * by default unless requested using gnutls_range_send_message().
+ *
  **/
 void
 gnutls_record_disable_padding (gnutls_session_t session)
 {
-  session->internals.priorities.no_padding = 1;
+  return;
+}
+
+/**
+ * gnutls_record_set_max_empty_records:
+ * @session: is a #gnutls_session_t structure.
+ * @i: is the desired value of maximum empty records that can be accepted in a row.
+ *
+ * Used to set the maximum number of empty fragments that can be accepted
+ * in a row. Accepting many empty fragments is useful for receiving length-hidden
+ * content, where empty fragments filled with pad are sent to hide the real
+ * length of a message. However, a malicious peer could send empty fragments to
+ * mount a DoS attack, so as a safety measure, a maximum number of empty fragments
+ * is accepted by default. If you know your application must accept a given number
+ * of empty fragments in a row, you can use this function to set the desired value.
+ **/
+void
+gnutls_record_set_max_empty_records (gnutls_session_t session, const unsigned int i)
+{
+  session->internals.priorities.max_empty_records = i;
 }
 
 /**
@@ -218,7 +241,8 @@ gnutls_bye (gnutls_session_t session, gnutls_close_request_t how)
         {
           do
             {
-              ret = _gnutls_recv_int (session, GNUTLS_ALERT, -1, NULL, 0, NULL);
+              ret = _gnutls_recv_int (session, GNUTLS_ALERT, -1, NULL, 0, NULL, 
+                                      session->internals.record_timeout_ms);
             }
           while (ret == GNUTLS_E_GOT_APPLICATION_DATA);
 
@@ -324,10 +348,10 @@ sequence_increment (gnutls_session_t session,
  *
  */
 ssize_t
-_gnutls_send_int (gnutls_session_t session, content_type_t type,
-                  gnutls_handshake_description_t htype,
-                  unsigned int epoch_rel, const void *_data,
-                  size_t data_size, unsigned int mflags)
+_gnutls_send_tlen_int (gnutls_session_t session, content_type_t type,
+		gnutls_handshake_description_t htype,
+		unsigned int epoch_rel, const void *_data,
+		size_t data_size, size_t target_length, unsigned int mflags)
 {
   mbuffer_st *bufel;
   ssize_t cipher_size;
@@ -380,17 +404,15 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
     memcpy(&headers[3], &record_state->sequence_number.i, 8);
 
   _gnutls_record_log
-    ("REC[%p]: Preparing Packet %s(%d) with length: %d\n", session,
-     _gnutls_packet2str (type), type, (int) data_size);
+    ("REC[%p]: Preparing Packet %s(%d) with length: %d and target length: %d\n", session,
+     _gnutls_packet2str (type), type, (int) data_size, (int) target_length);
 
-  if (data_size > MAX_RECORD_SEND_SIZE(session))
+  if (data_size > MAX_USER_SEND_SIZE(session))
     {
       if (IS_DTLS(session))
-        {
-          gnutls_assert ();
-          return GNUTLS_E_LARGE_PACKET;
-        }
-      send_data_size = MAX_RECORD_SEND_SIZE(session);
+        return gnutls_assert_val(GNUTLS_E_LARGE_PACKET);
+
+      send_data_size = MAX_USER_SEND_SIZE(session);
     }
   else
     send_data_size = data_size;
@@ -412,14 +434,14 @@ _gnutls_send_int (gnutls_session_t session, content_type_t type,
     {
       /* now proceed to packet encryption
        */
-      cipher_size = send_data_size + MAX_RECORD_OVERHEAD + CIPHER_SLACK_SIZE;
-      bufel = _mbuffer_alloc (cipher_size, cipher_size);
+      cipher_size = MAX_RECORD_SEND_SIZE(session);
+      bufel = _mbuffer_alloc (0, cipher_size+CIPHER_SLACK_SIZE);
       if (bufel == NULL)
         return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
       ret =
         _gnutls_encrypt (session, headers, header_size, data,
-                         send_data_size, _mbuffer_get_udata_ptr (bufel),
+                         send_data_size, target_length, _mbuffer_get_udata_ptr (bufel),
                          cipher_size, type, record_params);
       if (ret <= 0)
         {
@@ -485,6 +507,7 @@ check_recv_type (gnutls_session_t session, content_type_t recv_type)
     case GNUTLS_CHANGE_CIPHER_SPEC:
     case GNUTLS_ALERT:
     case GNUTLS_HANDSHAKE:
+    case GNUTLS_HEARTBEAT:
     case GNUTLS_APPLICATION_DATA:
       return 0;
     default:
@@ -650,6 +673,13 @@ record_add_to_buffers (gnutls_session_t session,
           _gnutls_record_buffer_put (session, recv->type, seq, bufel);
 
           break;
+
+#ifdef ENABLE_HEARTBEAT
+        case GNUTLS_HEARTBEAT:
+	    ret = _gnutls_heartbeat_handle (session, bufel);
+	    goto cleanup;
+#endif
+
         case GNUTLS_APPLICATION_DATA:
           if (session->internals.initial_negotiation_completed == 0)
             {
@@ -683,6 +713,7 @@ record_add_to_buffers (gnutls_session_t session,
             }
 
           break;
+
         case GNUTLS_HANDSHAKE:
           /* In DTLS we might receive a handshake replay from the peer to indicate
            * the our last TLS handshake messages were not received.
@@ -694,7 +725,7 @@ record_add_to_buffers (gnutls_session_t session,
                   ret = gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
                   goto unexpected_packet;
                 }
-                
+
               if (_dtls_is_async(session) && _dtls_async_timer_active(session))
                 {
                   if (session->security_parameters.entity == GNUTLS_SERVER &&
@@ -705,9 +736,11 @@ record_add_to_buffers (gnutls_session_t session,
                     }
                   else
                     {
+                      session->internals.recv_state = RECV_STATE_DTLS_RETRANSMIT;
                       ret = _dtls_retransmit(session);
                       if (ret == 0) 
                         {
+                          session->internals.recv_state = RECV_STATE_0;
                           ret = gnutls_assert_val(GNUTLS_E_AGAIN);
                           goto unexpected_packet;
                         }
@@ -747,8 +780,8 @@ record_add_to_buffers (gnutls_session_t session,
         default:
 
           _gnutls_record_log
-            ("REC[%p]: Received Unknown packet %d expecting %d\n",
-             session, recv->type, type);
+            ("REC[%p]: Received unexpected packet %d (%s) expecting %d (%s)\n",
+             session, recv->type, _gnutls_packet2str(recv->type), type, _gnutls_packet2str(type));
 
           gnutls_assert ();
           ret = GNUTLS_E_UNEXPECTED_PACKET;
@@ -863,7 +896,7 @@ record_read_headers (gnutls_session_t session,
 static int recv_headers( gnutls_session_t session, content_type_t type, 
                          gnutls_handshake_description_t htype, 
                          struct tls_record_st* record,
-                         unsigned int ms)
+                         unsigned int *ms)
 {
 int ret;
 gnutls_datum_t raw; /* raw headers */
@@ -871,8 +904,9 @@ gnutls_datum_t raw; /* raw headers */
    */
   record->header_size = record->packet_size = RECORD_HEADER_SIZE(session);
 
-  if ((ret =
-       _gnutls_io_read_buffered (session, record->header_size, -1, ms)) != record->header_size)
+  ret =
+       _gnutls_io_read_buffered (session, record->header_size, -1, ms);
+  if (ret != record->header_size)
     {
       if (ret < 0 && gnutls_error_is_fatal (ret) == 0)
         return ret;
@@ -939,9 +973,9 @@ gnutls_datum_t raw; /* raw headers */
   return 0;
 }
 
-#define MAX_EMPTY_PACKETS_SEQUENCE 4
-
-/* This will receive record layer packets and add them to 
+/* @ms: is the number of milliseconds to wait for data. Use zero for indefinite.
+ *
+ * This will receive record layer packets and add them to 
  * application_data_buffer and handshake_data_buffer.
  *
  * If the htype is not -1 then handshake timeouts
@@ -949,21 +983,21 @@ gnutls_datum_t raw; /* raw headers */
  */
 ssize_t
 _gnutls_recv_in_buffers (gnutls_session_t session, content_type_t type,
-                         gnutls_handshake_description_t htype)
+                         gnutls_handshake_description_t htype, unsigned int ms)
 {
   uint64 *packet_sequence;
-  uint8_t *ciphertext;
+  gnutls_datum_t ciphertext;
   mbuffer_st* bufel = NULL, *decrypted = NULL;
+  gnutls_datum_t t;
   int ret;
-  int empty_packet = 0;
+  unsigned int empty_fragments = 0;
   record_parameters_st *record_params;
   record_state_st *record_state;
   struct tls_record_st record;
-  time_t now, tleft = 0;
 
 begin:
 
-  if (empty_packet > MAX_EMPTY_PACKETS_SEQUENCE)
+  if (empty_fragments > session->internals.priorities.max_empty_records)
     {
       gnutls_assert ();
       return GNUTLS_E_TOO_MANY_EMPTY_PACKETS;
@@ -992,17 +1026,8 @@ begin:
 
   record_state = &record_params->read;
 
-  if (htype != (unsigned)-1 && session->internals.handshake_endtime > 0)
-    {
-      now = gnutls_time(0);
-      if (now < session->internals.handshake_endtime)
-        tleft = (session->internals.handshake_endtime - now) * 1000;
-      else
-        return gnutls_assert_val(GNUTLS_E_TIMEDOUT);
-    }
-
   /* receive headers */
-  ret = recv_headers(session, type, htype, &record, tleft);
+  ret = recv_headers(session, type, htype, &record, &ms);
   if (ret < 0)
     {
       ret = gnutls_assert_val_fatal(ret);
@@ -1014,20 +1039,11 @@ begin:
   else
     packet_sequence = &record_state->sequence_number;
 
-  if (htype != (unsigned)-1 && session->internals.handshake_endtime > 0)
-    {
-      now = gnutls_time(0);
-      if (now < session->internals.handshake_endtime)
-        tleft = (session->internals.handshake_endtime - now) * 1000;
-      else
-        return gnutls_assert_val(GNUTLS_E_TIMEDOUT);
-    }
-
   /* Read the packet data and insert it to record_recv_buffer.
    */
   ret =
        _gnutls_io_read_buffered (session, record.packet_size,
-                                 record.type, tleft);
+                                 record.type, &ms);
   if (ret != record.packet_size)
     {
       gnutls_assert();
@@ -1053,13 +1069,15 @@ begin:
   if (decrypted == NULL)
     return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
-  ciphertext = (uint8_t*)_mbuffer_get_udata_ptr(bufel) + record.header_size;
+  ciphertext.data = (uint8_t*)_mbuffer_get_udata_ptr(bufel) + record.header_size;
+  ciphertext.size = record.length;
 
   /* decrypt the data we got. 
    */
+  t.data = _mbuffer_get_udata_ptr(decrypted);
+  t.size = _mbuffer_get_udata_size(decrypted);
   ret =
-    _gnutls_decrypt (session, ciphertext, record.length, 
-        _mbuffer_get_udata_ptr(decrypted), _mbuffer_get_udata_size(decrypted),
+    _gnutls_decrypt (session, &ciphertext, &t,
 		     record.type, record_params, packet_sequence);
   if (ret >= 0) _mbuffer_set_udata_size(decrypted, ret);
 
@@ -1120,7 +1138,7 @@ begin:
   if (_mbuffer_get_udata_size(decrypted) == 0)
     {
       _mbuffer_xfree(&decrypted);
-      empty_packet++;
+      empty_fragments++;
       goto begin;
     }
 
@@ -1167,20 +1185,22 @@ cleanup:
   return ret;
 
 recv_error:
-  if (ret < 0 && gnutls_error_is_fatal (ret) == 0)
+  if (ret < 0 && (gnutls_error_is_fatal (ret) == 0 || ret == GNUTLS_E_TIMEDOUT))
     return ret;
 
+  if (type == GNUTLS_ALERT) /* we were expecting close notify */
+    {
+      session_invalidate (session);
+      gnutls_assert ();
+      return 0;             
+    }
+
   if (IS_DTLS(session))
     {
       goto discard;
     }
 
   session_invalidate (session);
-  if (type == GNUTLS_ALERT) /* we were expecting close notify */
-    {
-      gnutls_assert ();
-      return 0;             
-    }
   session_unresumable (session);
 
   if (ret == 0)
@@ -1199,14 +1219,13 @@ recv_error:
 ssize_t
 _gnutls_recv_int (gnutls_session_t session, content_type_t type,
                   gnutls_handshake_description_t htype,
-                  uint8_t * data, size_t data_size, void* seq)
+                  uint8_t * data, size_t data_size, void* seq,
+                  unsigned int ms)
 {
   int ret;
 
-  if (type != GNUTLS_ALERT && (data_size == 0 || data == NULL))
-    {
-      return GNUTLS_E_INVALID_REQUEST;
-    }
+  if ((type != GNUTLS_ALERT && type != GNUTLS_HEARTBEAT) && (data_size == 0 || data == NULL))
+    return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
   if (session->internals.read_eof != 0)
     {
@@ -1220,24 +1239,35 @@ _gnutls_recv_int (gnutls_session_t session, content_type_t type,
       gnutls_assert ();
       return GNUTLS_E_INVALID_SESSION;
     }
+    
+  switch(session->internals.recv_state)
+    {
+      case RECV_STATE_DTLS_RETRANSMIT:
+        ret = _dtls_retransmit(session);
+        if (ret < 0)
+          return gnutls_assert_val(ret);
+        
+        session->internals.recv_state = RECV_STATE_0;
+      case RECV_STATE_0:
 
-  _dtls_async_timer_check(session);
-  
-  /* If we have enough data in the cache do not bother receiving
-   * a new packet. (in order to flush the cache)
-   */
-  ret = check_buffers (session, type, data, data_size, seq);
-  if (ret != 0)
-    return ret;
+        _dtls_async_timer_check(session);
+        /* If we have enough data in the cache do not bother receiving
+         * a new packet. (in order to flush the cache)
+         */ 
+        ret = check_buffers (session, type, data, data_size, seq);
+        if (ret != 0)
+          return ret;
 
-  ret = _gnutls_recv_in_buffers(session, type, htype);
-  if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
-    return gnutls_assert_val(ret);
+        ret = _gnutls_recv_in_buffers(session, type, htype, ms);
+        if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
+          return gnutls_assert_val(ret);
 
-  return check_buffers (session, type, data, data_size, seq);
+        return check_buffers (session, type, data, data_size, seq);
+      default:
+        return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
+    }
 }
 
-
 /**
  * gnutls_record_send:
  * @session: is a #gnutls_session_t structure.
@@ -1269,9 +1299,96 @@ ssize_t
 gnutls_record_send (gnutls_session_t session, const void *data,
                     size_t data_size)
 {
-  return _gnutls_send_int (session, GNUTLS_APPLICATION_DATA, -1,
-                           EPOCH_WRITE_CURRENT, data, data_size,
-                           MBUFFER_FLUSH);
+  if (session->internals.record_flush_mode == RECORD_FLUSH)
+    {
+      return _gnutls_send_int (session, GNUTLS_APPLICATION_DATA, -1,
+                               EPOCH_WRITE_CURRENT, data, data_size,
+                               MBUFFER_FLUSH);
+    }
+  else /* GNUTLS_CORKED */
+    {
+      int ret;
+
+      ret = _gnutls_buffer_append_data(&session->internals.record_presend_buffer, data, data_size);
+      if (ret < 0)
+        return gnutls_assert_val(ret);
+      
+      return data_size;
+    }
+}
+
+/**
+ * gnutls_cork:
+ * @session: is a #gnutls_session_t structure.
+ *
+ * If called gnutls_record_send() will no longer send partial records.
+ * All queued records will be sent when gnutls_uncork() is called, or
+ * when the maximum record size is reached.
+ *
+ * Since: 3.1.7
+ **/
+void
+gnutls_record_cork (gnutls_session_t session)
+{
+  session->internals.record_flush_mode = RECORD_CORKED;
+}
+
+/**
+ * gnutls_uncork:
+ * @session: is a #gnutls_session_t structure.
+ * @flags: Could be zero or %GNUTLS_RECORD_WAIT
+ *
+ * This resets the effect of gnutls_cork(), and flushes any pending
+ * data. If the %GNUTLS_RECORD_WAIT flag is specified then this
+ * function will block until the data is sent or a fatal error
+ * occurs (i.e., the function will retry on %GNUTLS_E_AGAIN and
+ * %GNUTLS_E_INTERRUPTED).
+ *
+ * Returns: On success the number of transmitted data is returned, or 
+ * otherwise a negative error code. 
+ *
+ * Since: 3.1.7
+ **/
+int
+gnutls_record_uncork (gnutls_session_t session, unsigned int flags)
+{
+int ret;
+ssize_t total = 0;
+
+  if (session->internals.record_flush_mode == RECORD_FLUSH)
+    return 0; /* nothing to be done */
+
+  session->internals.record_flush_mode = RECORD_FLUSH;
+
+  while(session->internals.record_presend_buffer.length > 0)
+    {
+      if (flags == GNUTLS_RECORD_WAIT)
+        {
+          do
+            {
+              ret = gnutls_record_send(session, session->internals.record_presend_buffer.data, 
+                                       session->internals.record_presend_buffer.length);
+            }
+          while (ret < 0 && gnutls_error_is_fatal(ret) == 0);
+        }
+      else
+        {
+          ret = gnutls_record_send(session, session->internals.record_presend_buffer.data, 
+                                   session->internals.record_presend_buffer.length);
+        }
+      if (ret < 0)
+        goto fail;
+
+      session->internals.record_presend_buffer.data += ret;
+      session->internals.record_presend_buffer.length -= ret;
+      total += ret;
+    }
+
+  return total;
+
+fail:
+  session->internals.record_flush_mode = RECORD_CORKED;
+  return ret;
 }
 
 /**
@@ -1302,10 +1419,10 @@ gnutls_record_send (gnutls_session_t session, const void *data,
  * The number of bytes received might be less than the requested @data_size.
  **/
 ssize_t
-gnutls_record_recv (gnutls_session_t session, void *data, size_t data_size)
+gnutls_record_recv (gnutls_session_t session, void *data, size_t data_size) 
 {
-  return _gnutls_recv_int (session, GNUTLS_APPLICATION_DATA, -1, data,
-                           data_size, NULL);
+    return _gnutls_recv_int (session, GNUTLS_APPLICATION_DATA, -1, data, data_size, 
+                             NULL, session->internals.record_timeout_ms);
 }
 
 /**
@@ -1333,5 +1450,27 @@ gnutls_record_recv_seq (gnutls_session_t session, void *data, size_t data_size,
   unsigned char *seq)
 {
   return _gnutls_recv_int (session, GNUTLS_APPLICATION_DATA, -1, data,
-                           data_size, seq);
+                           data_size, seq, session->internals.record_timeout_ms);
+}
+
+/**
+ * gnutls_record_set_timeout:
+ * @session: is a #gnutls_session_t structure.
+ * @ms: is a timeout value in milliseconds
+ *
+ * This function sets the receive timeout for the record layer
+ * to the provided value. Use an @ms value of zero to disable
+ * timeout (the default).
+ *
+ * Note that in order for the timeout to be enforced
+ * gnutls_transport_set_pull_timeout_function() must be set
+ * (it is set by default in most systems).
+ *
+ * Since: 3.1.7
+ *
+ **/
+void
+gnutls_record_set_timeout (gnutls_session_t session, unsigned int ms)
+{
+  session->internals.record_timeout_ms = ms;
 }
