@@ -271,18 +271,22 @@ generate_certificate (gnutls_privkey_t * ret_key,
             error (EXIT_FAILURE, 0, "set_proxy_dn: %s",
                    gnutls_strerror (result));
 
+          get_dn_crt_set (crt);
           get_cn_crt_set (crt);
         }
       else
         {
-          get_country_crt_set (crt);
-          get_organization_crt_set (crt);
+          get_dn_crt_set (crt);
+
+          get_cn_crt_set (crt);
+          get_uid_crt_set (crt);
           get_unit_crt_set (crt);
+          get_organization_crt_set (crt);
           get_locality_crt_set (crt);
           get_state_crt_set (crt);
-          get_cn_crt_set (crt);
+          get_country_crt_set (crt);
           get_dc_set (TYPE_CRT, crt);
-          get_uid_crt_set (crt);
+
           get_oid_crt_set (crt);
           get_key_purpose_set (crt);
 
@@ -924,7 +928,7 @@ cmd_parser (int argc, char **argv)
     req_key_type = GNUTLS_PK_ECC;
   else
     req_key_type = GNUTLS_PK_RSA;
-
+    
   default_dig = GNUTLS_DIG_UNKNOWN;
   if (HAVE_OPT(HASH))
     {
@@ -976,6 +980,8 @@ cmd_parser (int argc, char **argv)
   
   if (HAVE_OPT(VERBOSE))
     cinfo.verbose = 1;
+    
+  cinfo.cprint = HAVE_OPT(CPRINT);
   
   if (HAVE_OPT(LOAD_PRIVKEY))
     cinfo.privkey = OPT_ARG(LOAD_PRIVKEY);
@@ -1051,7 +1057,7 @@ cmd_parser (int argc, char **argv)
   else if (HAVE_OPT(CERTIFICATE_INFO))
     certificate_info (0, &cinfo);
   else if (HAVE_OPT(DH_INFO))
-    dh_info (&cinfo);
+    dh_info (infile, outfile, &cinfo);
   else if (HAVE_OPT(CERTIFICATE_PUBKEY))
     certificate_info (1, &cinfo);
   else if (HAVE_OPT(KEY_INFO))
@@ -1063,9 +1069,9 @@ cmd_parser (int argc, char **argv)
   else if (HAVE_OPT(P12_INFO))
     pkcs12_info (&cinfo);
   else if (HAVE_OPT(GENERATE_DH_PARAMS))
-    generate_prime (1, &cinfo);
+    generate_prime (outfile, 1, &cinfo);
   else if (HAVE_OPT(GET_DH_PARAMS))
-    generate_prime (0, &cinfo);
+    generate_prime (outfile, 0, &cinfo);
   else if (HAVE_OPT(CRL_INFO))
     crl_info ();
   else if (HAVE_OPT(P7_INFO))
@@ -1296,7 +1302,7 @@ pgp_privkey_info (void)
             fprintf (stderr, "Error in key RSA data export: %s\n",
                      gnutls_strerror (ret));
           else
-            print_rsa_pkey (outfile, &m, &e, &d, &p, &q, &u, NULL, NULL);
+            print_rsa_pkey (outfile, &m, &e, &d, &p, &q, &u, NULL, NULL, HAVE_OPT(CPRINT));
 
           bits = m.size * 8;
         }
@@ -1315,7 +1321,7 @@ pgp_privkey_info (void)
             fprintf (stderr, "Error in key DSA data export: %s\n",
                      gnutls_strerror (ret));
           else
-            print_dsa_pkey (outfile, &x, &y, &p, &q, &g);
+            print_dsa_pkey (outfile, &x, &y, &p, &q, &g, HAVE_OPT(CPRINT));
             
           bits = y.size * 8;
         }
@@ -1609,7 +1615,7 @@ const char *cprint;
                  gnutls_strerror (ret));
       else
         {
-          print_rsa_pkey (outfile, &m, &e, &d, &p, &q, &u, &exp1, &exp2);
+          print_rsa_pkey (outfile, &m, &e, &d, &p, &q, &u, &exp1, &exp2, HAVE_OPT(CPRINT));
 
           gnutls_free (m.data);
           gnutls_free (e.data);
@@ -1631,7 +1637,7 @@ const char *cprint;
                  gnutls_strerror (ret));
       else
         {
-          print_dsa_pkey (outfile, &x, &y, &p, &q, &g);
+          print_dsa_pkey (outfile, &x, &y, &p, &q, &g, HAVE_OPT(CPRINT));
 
           gnutls_free (x.data);
           gnutls_free (y.data);
@@ -1651,7 +1657,7 @@ const char *cprint;
                  gnutls_strerror (ret));
       else
         {
-          print_ecc_pkey (outfile, curve, &k, &x, &y);
+          print_ecc_pkey (outfile, curve, &k, &x, &y, HAVE_OPT(CPRINT));
 
           gnutls_free (x.data);
           gnutls_free (y.data);
@@ -1778,12 +1784,15 @@ generate_request (common_info_st * cinfo)
 
   /* Set the DN.
    */
-  get_country_crq_set (crq);
-  get_organization_crq_set (crq);
+  get_dn_crq_set (crq);
+
+  get_cn_crq_set (crq);
   get_unit_crq_set (crq);
+  get_organization_crq_set (crq);
   get_locality_crq_set (crq);
   get_state_crq_set (crq);
-  get_cn_crq_set (crq);
+  get_country_crq_set (crq);
+
   get_dc_set (TYPE_CRQ, crq);
   get_uid_crq_set (crq);
   get_oid_crq_set (crq);
